﻿Public Class AnalyzeTS

    Public Delegate Sub NewTSMEssageHandler(ByVal Message As String)
    Public Event OnNewTSMessage As NewTSMEssageHandler
    Public Delegate Sub NewPATMEssageHandler(ByVal Message As String)
    Public Event OnNewPATMessage As NewPATMEssageHandler
    Public Delegate Sub NewPMTMEssageHandler(ByVal Message As String)
    Public Event OnNewPMTMessage As NewPMTMEssageHandler

    'Public Delegate Sub NewPATHandler()
    'Public Event OnNewPAT As NewPATHandler

    Public Class InfoPID
        Public Received As Int32 = 0
        Public Unscrambled As Int32 = 0
        Public Scrambled As Int32 = 0
        Public HasAdaptationFields As Int32 = 0
        Public HasPayload As Int32 = 0
        Public StreamPriority As Int32 = 0
        Public StreamError As Int32 = 0
        Public ContinuityError As Int32 = 0
        Public ContinuityCounter As Int32 = -1
        Public Type As String = "Unknown"
        Public Description As String = "Unknown"
        Public [Optional] As String = "None"
        Public Analysis As Int32 = 0
    End Class

    Public TSSynchro As Byte = &H47
    Public TSPacketSize As Int32 = 188
    Public TSReceived As Int32 = 0
    Public TSScrambled As Int32 = 0
    Public TSUnscrambled As Int32 = 0
    Public TSError As Int32 = 0
    Public TSPIDS As New Hashtable()

    Public PATVersion As Int32 = -1
    Public PMTVersion As New Hashtable()

    Public Sub New()
    End Sub

    Public Sub NewPacket(ByVal Packet As Byte())
        ' TODO : Does not handle the possibility to have duplicate and consecutive packets (same packets, with updated PCR)
        ' New Packet
        TSReceived += 1
        ' Do some checks - Packet size
        If Packet.Length <> TSPacketSize Then
            TSError += 1
            SendTSMessage("ERROR: Receive a " + Packet.Length.ToString + " bytes size packet (expected size is " + TSPacketSize.ToString + " bytes)")
            Return
        End If
        ' Do some checks - Synchro
        If Packet(0) <> TSSynchro Then
            TSError += 1
            SendTSMessage("ERROR: Missing the synchro byte 0x" + TSSynchro.ToString("X2") + " (found 0x" + Packet(0).ToString("X2") + ")")
            Return
        End If
        ' General TS information extraction - 2 first bytes after synchro byte
        Dim MPEG_transport_error_indicator As Boolean = False
        If ((Packet(1) And &H80) = &H80) Then MPEG_transport_error_indicator = True
        Dim MPEG_payload_unit_start_indicator As Boolean = False
        If ((Packet(1) And &H40) = &H40) Then MPEG_payload_unit_start_indicator = True
        Dim MPEG_transport_priority As Boolean = False
        If ((Packet(1) And &H20) = &H20) Then MPEG_transport_priority = True
        Dim MPEG_PID As Int32 = (CInt(Packet(1) And &H1F) << 8) + CInt(Packet(2))
        Dim MPEG_transport_scrambling_control As Int32 = (CInt(Packet(3) And &HC0) >> 6)
        Dim MPEG_adaptation_field_control As Int32 = (CInt(Packet(3) And &H30) >> 4)
        Dim MPEG_continuity_counter As Int32 = CInt(Packet(3) And &HF)
        ' Create PID info if needed
        If TSPIDS.ContainsKey(MPEG_PID) = False Then
            Dim Info As New InfoPID()
            TSPIDS.Add(MPEG_PID, Info)
            SendTSMessage("INFO: Adding new PID=" + MPEG_PID.ToString)
        End If
        ' Adaptation field handling
        Dim HasAdaptationField As Boolean = False
        Dim HasPayload As Boolean = False
        If (MPEG_adaptation_field_control = 2 Or MPEG_adaptation_field_control = 3) Then HasAdaptationField = True
        If (MPEG_adaptation_field_control = 1 Or MPEG_adaptation_field_control = 3) Then HasPayload = True
        Dim AdaptationFieldLength As Int32 = 0 ' Real total length of the Adapatation Field section
        Dim MPEG_adaptation_field_length As Int32 = 0
        If (HasAdaptationField = True) Then
            MPEG_adaptation_field_length = CInt(Packet(4))
            AdaptationFieldLength = MPEG_adaptation_field_length + 1
            If ((MPEG_adaptation_field_control = 2 And AdaptationFieldLength <= 183) Or (MPEG_adaptation_field_control = 3 And AdaptationFieldLength <= 182)) Then
                ' TODO : Adapatation Field analysis
            End If
        End If
        ' Payload fiels handling
        Dim PayloadLength As Int32 = 0
        Dim PayloadPointer As Int32 = 0
        If (HasPayload = True) Then
            PayloadLength = 184 - AdaptationFieldLength
            PayloadPointer = 4 + AdaptationFieldLength
            If (PayloadLength > 0 And PayloadPointer + PayloadLength = 188 And MPEG_PID <> &H1FFF) Then
                ' Payload Analysis
                PayloadAnalysis(MPEG_PID, MPEG_payload_unit_start_indicator, MPEG_continuity_counter, PayloadPointer, PayloadLength, Packet)
            End If
        End If
        ' Speical case: Null packet handling
        If MPEG_PID = &H1FFF Then
            DirectCast(TSPIDS(MPEG_PID), InfoPID).Type = "NUL"
            DirectCast(TSPIDS(MPEG_PID), InfoPID).Description = "Nul packet"
        End If
        ' Error/Warning messages
        If (MPEG_transport_error_indicator = True) Then SendTSMessage("WARNING: TS Transport Error Indicator is set for packet PID=" + MPEG_PID.ToString)
        If (MPEG_transport_priority = True) Then SendTSMessage("INFO: TS Transport Priority is set for packet PID=" + MPEG_PID.ToString)
        If (MPEG_adaptation_field_control = 0) Then SendTSMessage("WARNING: TS Adaptation Field Control is set to 0 for packet PID=" + MPEG_PID.ToString)
        If (MPEG_adaptation_field_control = 2 And MPEG_adaptation_field_length > 183) Then SendTSMessage("ERROR: TS Adaptation Field Control is set to 0x2 and Adaptation Filed size is over 183 bytes (" + AdaptationFieldLength.ToString() + " bytes) for packet PID=" + MPEG_PID.ToString)
        If (MPEG_adaptation_field_control = 3 And MPEG_adaptation_field_length > 182) Then SendTSMessage("ERROR: TS Adaptation Field Control is set to 0x3 and Adaptation Filed size is over 182 bytes (" + AdaptationFieldLength.ToString() + " bytes) for packet PID=" + MPEG_PID.ToString)
        If (PayloadLength < 0) Then SendTSMessage("ERROR: Payload size is negative (" + PayloadLength.ToString + ") for packet PID=" + MPEG_PID.ToString)
        If (HasPayload = True And PayloadPointer + PayloadLength <> 188) Then SendTSMessage("ERROR: Payload size is not coherent with its pointeur (Pointeur=" + PayloadPointer.ToString + " Payload=" + PayloadLength.ToString + " bytes) for packet PID=" + MPEG_PID.ToString)
        If (MPEG_PID = &H1FFF And MPEG_transport_scrambling_control <> 0) Then SendTSMessage("WARNING: Transport Srambling Control must be 0 (null packet) for packet PID=" + MPEG_PID.ToString)
        If (MPEG_PID = &H1FFF And MPEG_adaptation_field_control <> 1) Then SendTSMessage("WARNING: Adaptation Field Control must be 0x1 for null packet (0x" + MPEG_adaptation_field_control.ToString + " received) for packet PID=" + MPEG_PID.ToString)
        ' Statistics
        If MPEG_transport_scrambling_control = 0 Then
            TSUnscrambled += 1
            DirectCast(TSPIDS(MPEG_PID), InfoPID).Unscrambled += 1
        Else
            TSScrambled += 1
            DirectCast(TSPIDS(MPEG_PID), InfoPID).Scrambled += 1
        End If
        DirectCast(TSPIDS(MPEG_PID), InfoPID).Received += 1
        If HasAdaptationField = True Then DirectCast(TSPIDS(MPEG_PID), InfoPID).HasAdaptationFields += 1
        If HasPayload = True Then DirectCast(TSPIDS(MPEG_PID), InfoPID).HasPayload += 1
        If MPEG_transport_error_indicator = True Then DirectCast(TSPIDS(MPEG_PID), InfoPID).StreamError += 1
        If MPEG_transport_priority = True Then DirectCast(TSPIDS(MPEG_PID), InfoPID).StreamPriority += 1
    End Sub

    Public Sub SendTSMessage(ByVal Message As String)
        RaiseEvent OnNewTSMessage(Message)
    End Sub

    Public Sub SendPATMessage(ByVal Message As String)
        RaiseEvent OnNewPATMessage(Message)
    End Sub

    Public Sub SendPMTMessage(ByVal Message As String)
        RaiseEvent OnNewPMTMessage(Message)
    End Sub

    Public Sub PayloadAnalysis(ByVal MPEG_PID As Int32, ByVal MPEG_payload_unit_start_indicator As Boolean, ByVal MPEG_continuity_counter As Int32, ByVal PayloadPointer As Int32, ByVal PayloadLength As Int32, ByVal Packet As Byte())

        ' Check continuity
        Dim Info As InfoPID = DirectCast(TSPIDS(MPEG_PID), InfoPID)
        If Info.ContinuityCounter <> -1 Then
            If (MPEG_continuity_counter >= 1 And MPEG_continuity_counter <= 15 And MPEG_continuity_counter = Info.ContinuityCounter + 1) Or (MPEG_continuity_counter = 0 And Info.ContinuityCounter = 15) Then
                ' OK
            Else
                ' Continuity error
                Info.ContinuityError += 1
                ' Stop current analysis
                Info.Analysis = 0
            End If
        End If
        Info.ContinuityCounter = MPEG_continuity_counter

        ' Packet type identification
        If MPEG_payload_unit_start_indicator = True Then

            ' Test if PES stream - Search for PES packet_start_code_prefix=0x000001
            If PayloadLength > 4 AndAlso Packet(PayloadPointer + 0) = 0 AndAlso Packet(PayloadPointer + 1) = 0 AndAlso Packet(PayloadPointer + 2) = 1 Then
                Info.Type = "PES"
                Info.Description = StreamDescription(Packet(PayloadPointer + 3))
            Else
                ' Check if we have a PSI stream
                Dim MPEG_pointer_field As Byte = Packet(PayloadPointer + 0)
                ' Check the pointer limit
                If MPEG_pointer_field + 2 <= PayloadLength Then

                    ' Get the table id
                    Dim MPEG_table_id As Byte = Packet(PayloadPointer + 1 + MPEG_pointer_field)

                    ' Reset in-depth analysis
                    Info.Analysis = 0

                    ' --- PSI Tables ---

                    ' PSI-PAT
                    If MPEG_table_id = 0 Then
                        If MPEG_PID = 0 Then
                            ' PAT
                            Info.Type = "PSI"
                            Info.Description = "Program Association Table (PAT)"
                            Info.Analysis = 1
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (PAT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If
                    ' PSI-CAT
                    If MPEG_table_id = 1 Then
                        If MPEG_PID = 1 Then
                            ' PAT
                            Info.Type = "PSI"
                            Info.Description = "Conditional Association Table (CAT)"
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (CAT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If
                    ' PSI-PMT
                    If MPEG_table_id = 2 Then
                        If MPEG_PID >= &H10 And MPEG_PID <= &H1FFE Then
                            ' PMT
                            Info.Type = "PSI"
                            Info.Description = "Program Map Table (PMT)"
                            Info.Analysis = 2
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (PMT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If
                    ' PSI-TSDT
                    If MPEG_table_id = 3 Then
                        If MPEG_PID = 2 Then
                            ' TSDT
                            Info.Type = "PSI"
                            Info.Description = "Transport Stream Description Table (TSDT)"
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (TSDT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' --- DVB Mandatory Tables ---

                    ' DVB-Mandatory-NIT-ActualNetworkStream
                    If MPEG_table_id = &H40 Then
                        If MPEG_PID = &H10 Then
                            Info.Type = "DVB"
                            Info.Description = "Network Information Table (NIT)"
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (NIT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Mandatory-SDT-ActualNetworkStream
                    If MPEG_table_id = &H42 Then
                        If MPEG_PID = &H11 Then
                            Info.Type = "DVB"
                            Info.Description = "Service Description Table (SDT)"
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (SDT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Mandatory-EIT-ActualNetworkStream-PresentFollowing
                    If MPEG_table_id = &H4E Then
                        If MPEG_PID = &H12 Then
                            Info.Type = "DVB"
                            Info.Description = "Event Information Table (EIT)"
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (EIT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Mandatory-TDT
                    If MPEG_table_id = &H70 Then
                        If MPEG_PID = &H14 Then
                            Info.Type = "DVB"
                            Info.Description = "Time Date Table (TDT)"
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (TDT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' --- DVB Optional Tables ---

                    ' DVB-Optional-NIT-OtherTransportStream (0x10/0x41)
                    If MPEG_table_id = &H41 Then
                        If MPEG_PID = &H10 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-NIT-OtherTransportStream") < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-NIT-OtherTransportStream"
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-NIT-OtherTransportStream) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-BAT (0x11/0x4A)
                    If MPEG_table_id = &H4A Then
                        If MPEG_PID = &H11 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-BAT") < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-BAT"
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-BAT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-SDT-OtherTransportStream (0x11/0x46)
                    If MPEG_table_id = &H46 Then
                        If MPEG_PID = &H11 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-SDT-OtherTransportStream") < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-SDT-OtherTransportStream"
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-SDT-OtherTransportStream) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-EIT-ActualTransportStreamSchedule (0x12/0x50-0x5F)
                    If MPEG_table_id >= &H50 And MPEG_table_id <= &H5F Then
                        If MPEG_PID = &H12 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-EIT-ActualTransportStreamSchedule-" + MPEG_table_id.ToString("X2")) < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-EIT-ActualTransportStreamSchedule-" + MPEG_table_id.ToString("X2")
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-EIT-ActualTransportStreamSchedule-" + MPEG_table_id.ToString("X2") + ") found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-EIT-OtherTransportStreamPresentFollowing (0x12/0x4F)
                    If MPEG_table_id = &H4F Then
                        If MPEG_PID = &H12 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-EIT-OtherTransportStreamPresentFollowing") < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-EIT-OtherTransportStreamPresentFollowing"
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-EIT-OtherTransportStreamPresentFollowing) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-EIT-OtherTransportStreamSchedule (0x12/0x60-0x6F)
                    If MPEG_table_id >= &H60 And MPEG_table_id <= &H6F Then
                        If MPEG_PID = &H12 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-EIT-OtherTransportStreamSchedule-" + MPEG_table_id.ToString("X2")) < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-EIT-OtherTransportStreamSchedule-" + MPEG_table_id.ToString("X2")
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-EIT-OtherTransportStreamSchedule-" + MPEG_table_id.ToString("X2") + ") found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-RST (0x13/0x71)
                    If MPEG_table_id = &H71 Then
                        If MPEG_PID = &H13 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-RST") < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-RST"
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-RST) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-TOT (0x14/0x73)
                    If MPEG_table_id = &H73 Then
                        If MPEG_PID = &H14 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-TOT") < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-TOT"
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-TOT) found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                    ' DVB-Optional-ST (0x10-0x14/0x72)
                    If MPEG_table_id = &H72 Then
                        If MPEG_PID >= &H10 And MPEG_PID <= &H14 Then
                            If Info.Optional = "None" Then Info.Optional = ""
                            If Info.Optional.IndexOf("DVB-Optional-ST-" + MPEG_PID.ToString("X2")) < 0 Then
                                If Info.Optional <> "" Then Info.Optional += " "
                                Info.Optional += "DVB-Optional-ST-" + MPEG_PID.ToString("X2")
                            End If
                        Else
                            ' Problem
                            SendTSMessage("ERROR: table_id=" + MPEG_table_id.ToString + " (DVB-Optional-ST-" + MPEG_PID.ToString("X2") + " found for PID=" + MPEG_PID.ToString)
                        End If
                    End If

                Else
                    ' Pointer problem
                    SendTSMessage("ERROR: pointer_field is over payload data for packet PID=" + MPEG_PID.ToString)
                End If

            End If

            ' In-depth analysis of PAT
            If Info.Analysis = 1 Then
                AnalyzePAT(MPEG_payload_unit_start_indicator, PayloadPointer, PayloadLength, Packet)
            End If

            ' In-depth analysis of PMT
            If Info.Analysis = 2 Then
                AnalyzePMT(MPEG_PID, MPEG_payload_unit_start_indicator, PayloadPointer, PayloadLength, Packet)
            End If

        End If

    End Sub

    ' PAT analysis
    Public Sub AnalyzePAT(ByVal MPEG_payload_unit_start_indicator As Boolean, ByVal PayloadPointer As Int32, ByVal PayloadLength As Int32, ByVal Packet As Byte())

        ' PayloadPointer+1 = Index of "table_id" (PSI section)

        ' Test section_syntax_indicator
        If (CInt(Packet(PayloadPointer + 2)) And CInt(&HC0)) <> CInt(&H80) Then
            SendPATMessage("ERROR: Invalid section_syntax_indicator")
        End If

        ' Test if it's the begining of PSI 
        If MPEG_payload_unit_start_indicator = True Then
            ' Test if it's current version of PAT - current_next_indicator=1
            If (Packet(PayloadPointer + 6) And CByte(&H1)) = CByte(&H1) Then
                ' Get the PAT version 
                Dim PAT_Actual_Version As Byte = (Packet(PayloadPointer + 6) And CByte(&H3E)) >> 1
                ' New PAT version
                If PATVersion <> PAT_Actual_Version Then
                    ' Check PAT version
                    If PATVersion = -1 Then
                        ' First PAT
                        SendPATMessage("INFO: First PAT found with version " + PAT_Actual_Version.ToString)
                    Else
                        ' New PAT
                        SendPATMessage("INFO: Found new PAT version " + PAT_Actual_Version.ToString + ", last PAT version was " + PATVersion.ToString)
                        If (PAT_Actual_Version = PATVersion + 1) Or (PATVersion = 31 And PAT_Actual_Version = 0) Then
                            SendPATMessage("INFO: New version is consecutive to previous version")
                        Else
                            SendPATMessage("ERROR: New version isn't consecutive to previous version")
                        End If
                    End If
                    PATVersion = PAT_Actual_Version
                    ' Some extra information
                    SendPATMessage("INFO:  * transport_stream_id = " + CInt(Packet(PayloadPointer + 4) * 256 + Packet(PayloadPointer + 5)).ToString)
                    SendPATMessage("INFO:  * section_number      = " + CInt(Packet(PayloadPointer + 7)).ToString)
                    SendPATMessage("INFO:  * last_section_number = " + CInt(Packet(PayloadPointer + 8)).ToString)
                    Dim PAT_Section_Length As Int32 = CInt((Packet(PayloadPointer + 2) And CByte(&HF)) * 256 + Packet(PayloadPointer + 3))
                    SendPATMessage("INFO:  * section_length      = " + PAT_Section_Length.ToString)
                    If PAT_Section_Length > 1021 Then
                        SendPATMessage("ERROR: section_length shall not exceed 1021")
                    End If
                    SendPATMessage("INFO:  * associations        = " + CInt((PAT_Section_Length - 9) / 4).ToString)
                    ' List the first associations in the actual TS packet
                    Dim index As Int32 = PayloadPointer + 9
                    Dim number As Int32 = 0
                    Do While (number * 4 + 9 < PAT_Section_Length) And (index + 3 < 188)
                        If (Packet(index) = 0 And Packet(index + 1) = 0) Then
                            ' if program_number = 0 => NIT
                            SendPATMessage("INFO:    o network_PID=" + CInt((Packet(index + 2) And &H1F) * 256 + Packet(index + 3)).ToString)
                        Else
                            SendPATMessage("INFO:    o program_number=" + CInt(Packet(index) * 256 + Packet(index + 1)).ToString + " for program_map_PID=" + CInt((Packet(index + 2) And &H1F) * 256 + Packet(index + 3)).ToString)
                        End If
                        index += 4
                        number += 1
                    Loop
                    If number * 4 + 9 <> PAT_Section_Length Then
                        SendPATMessage("WARNING: Not all the associations were listed (only the ones from the first TS PAT packet)")
                    End If
                End If
            End If
        End If

    End Sub

    ' PMT analysis
    Public Sub AnalyzePMT(ByVal MPEG_PID As Int32, ByVal MPEG_payload_unit_start_indicator As Boolean, ByVal PayloadPointer As Int32, ByVal PayloadLength As Int32, ByVal Packet As Byte())

        ' PayloadPointer+1 = Index of "table_id" (PSI section)

        ' Test section_syntax_indicator
        If (CInt(Packet(PayloadPointer + 2)) And CInt(&HC0)) <> CInt(&H80) Then
            SendPMTMessage("ERROR: Invalid section_syntax_indicator for PID=" + MPEG_PID.ToString)
        End If

        ' Test if it's the begining of PSI 
        If MPEG_payload_unit_start_indicator = True Then
            ' Test if it's current version of PMT - current_next_indicator=1
            If (Packet(PayloadPointer + 6) And CByte(&H1)) = CByte(&H1) Then
                ' Get the PMT version 
                Dim PMT_ActualVersion As Byte = (Packet(PayloadPointer + 6) And CByte(&H3E)) >> 1
                ' Check if the PMT is already known
                Dim ShowExtraInfos As Boolean = False
                If PMTVersion.ContainsKey(MPEG_PID) = False Then
                    ' New PMT PID
                    SendPMTMessage("INFO: New PMT PID=" + MPEG_PID.ToString + " found with version " + PMT_ActualVersion.ToString)
                    PMTVersion.Add(MPEG_PID, PMT_ActualVersion)
                    ShowExtraInfos = True
                Else
                    ' Already known PID, we check if there is a change
                    Dim PMT_PreviousVersion As Byte = CByte(PMTVersion(MPEG_PID))
                    If PMT_ActualVersion <> PMT_PreviousVersion Then
                        ' New PMT version
                        SendPMTMessage("INFO: Found new PMT version " + PMT_ActualVersion.ToString + " for PID=" + MPEG_PID.ToString + ", last PAT version was " + PMT_PreviousVersion.ToString)
                        If (PMT_ActualVersion = PMT_PreviousVersion + 1) Or (PMT_PreviousVersion = 31 And PMT_ActualVersion = 0) Then
                            SendPMTMessage("INFO: New version is consecutive to previous version")
                        Else
                            SendPMTMessage("ERROR: New version isn't consecutive to previous version")
                        End If
                        ' Update the version
                        PMTVersion(MPEG_PID) = PMT_ActualVersion
                        ShowExtraInfos = True
                    End If
                End If
                ' Some extra information
                If ShowExtraInfos = True Then
                    SendPMTMessage("INFO:  * program_number      = " + CInt(Packet(PayloadPointer + 4) * 256 + Packet(PayloadPointer + 5)).ToString)
                    SendPMTMessage("INFO:  * section_number      = " + CInt(Packet(PayloadPointer + 7)).ToString)
                    SendPMTMessage("INFO:  * last_section_number = " + CInt(Packet(PayloadPointer + 8)).ToString)
                    Dim PMT_Section_Length As Int32 = CInt((Packet(PayloadPointer + 2) And CByte(&HF)) * 256 + Packet(PayloadPointer + 3))
                    SendPMTMessage("INFO:  * section_length      = " + PMT_Section_Length.ToString)
                    If PMT_Section_Length > 1021 Then
                        SendPMTMessage("ERROR: section_length shall not exceed 1021")
                    End If
                    SendPMTMessage("INFO:  * PCR_PID             = " + CInt((Packet(PayloadPointer + 9) And &H1F) * 256 + Packet(PayloadPointer + 10)).ToString)
                End If
            End If
        End If
    End Sub

    Public Function StreamDescription(ByVal StreamID As Int32) As String
        Dim Label As String = "Stream ID=0x" + StreamID.ToString("X2")
        Select Case StreamID
            Case &HBC
                Label += " (program_stream_map)"
            Case &HBD
                Label += " (private_stream_1)"
            Case &HBE
                Label += " (padding_stream)"
            Case &HBF
                Label += " (private_stream_2)"
            Case &HF0
                Label += " (ECM_stream)"
            Case &HF1
                Label += " (EMM_stream)"
            Case &HF2
                Label += " (DSMCC_stream)"
            Case &HF3
                Label += " (ISO/IEC_13522_stream)"
            Case &HF4
                Label += " (ITU_H.222.1_TypeA)"
            Case &HF5
                Label += " (ITU_H.222.1_TypeB)"
            Case &HF6
                Label += " (ITU_H.222.1_TypeC)"
            Case &HF7
                Label += " (ITU_H.222.1_TypeD)"
            Case &HF8
                Label += " (ITU_H.222.1_TypeE)"
            Case &HF9
                Label += " (ancillary_stream)"
            Case &HFA
                Label += " (SL-Packetized_stream)"
            Case &HFB
                Label += " (FlexMux_stream)"
            Case &HFF
                Label += " (program_stream_directory)"
        End Select
        If StreamID >= &HC0 And StreamID <= &HDF Then Label += " (Audio stream number #" + CInt(StreamID And &H1F).ToString + ")"
        If StreamID >= &HE0 And StreamID <= &HEF Then Label += " (Video stream number #" + CInt(StreamID And &HF).ToString + ")"
        If StreamID >= &HFC And StreamID <= &HFE Then Label += " (reserved data stream)"
        Return (Label)
    End Function

End Class
