﻿Imports System.Net
Imports System.Net.Sockets

Public Class Main

    Private Analyzer As AnalyzeTS = Nothing
    Private MessageTS As String = ""
    Private MessagePAT As String = ""
    Private MessagePMT As String = ""

    Private StreamInput As Int32 = 0
    Private StreamType As Int32 = 0
    Private StreamLocation As String = ""
    Private StreamAddress As IPEndPoint = Nothing

    Private DumpMode As Int32 = 0
    Private DumpFile As String = ""
    Private DumpMaxSize As Int64 = 0
    Private DumpMaxTime As Int64 = 0
    Private DumpActualFile As String = ""
    Private DumpTime As DateTime
    Private DumpSize As Int64 = 0
    Private DumpFileNumber As Int64 = 0
    Private DumpHandle As IO.FileStream = Nothing

    Private ListenUdp As UdpClient = Nothing

    Private Input_Source As String = ""
    Private Input_TSEstimatedSize As Int64 = 0
    Private Input_PacketsReceived As Int64 = 0
    Private Input_TSPacketsReceived As Int64 = 0
    Private Input_BadLength As Int64 = 0
    Private RTP_OutOfOrder As Int64 = 0
    Private RTP_BadHeader As Int64 = 0

    Private Previous_Input_TSPacketsReceived As Int64 = 0

    Private PreviousPIDs As Hashtable = Nothing

    Private Sub UpdateStats_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles UpdateStats.Tick
        UpdateStatsNow()
    End Sub

    Private Sub UpdateStatsNow()
        If IsNothing(Analyzer) = False Then

            ' TS Stats
            tbTSStats.Text = ""
            tbTSStats.Text += "Number of PIDs      : " + Analyzer.TSPIDS.Count.ToString + vbCrLf
            tbTSStats.Text += "Received packets    : " + Analyzer.TSReceived.ToString + vbCrLf
            tbTSStats.Text += "Unscrambled packets : " + Analyzer.TSUnscrambled.ToString + vbCrLf
            tbTSStats.Text += "Scrambled packets   : " + Analyzer.TSScrambled.ToString + vbCrLf
            tbTSStats.Text += "Error packets       : " + Analyzer.TSError.ToString + vbCrLf

            ' INPUT Stats
            tbRTPStats.Text = ""
            ' File
            If StreamInput = 0 Then
                tbRTPStats.Text += "TS estimated packets : " + Input_TSEstimatedSize.ToString + vbCrLf
                tbRTPStats.Text += "TS read packets      : " + Input_TSPacketsReceived.ToString + vbCrLf
                tbRTPStats.Text += "TS resynchronization : " + Input_BadLength.ToString + vbCrLf
            End If
            ' UDP Multicast
            If StreamInput = 1 Then
                tbRTPStats.Text += "UDP source       : " + Input_Source + vbCrLf
                tbRTPStats.Text += "UDP read packets : " + Input_PacketsReceived.ToString + vbCrLf
                tbRTPStats.Text += "UDP bad length   : " + Input_BadLength.ToString + vbCrLf
                tbRTPStats.Text += "TS bandwidth     : " + CDbl((Input_TSPacketsReceived - Previous_Input_TSPacketsReceived) * 188 * 8 / 1000000).ToString("0.000") + " mbps" + vbCrLf
                tbRTPStats.Text += "TS per UDP ratio : " + CDbl(Input_TSPacketsReceived / Input_PacketsReceived).ToString("0.0") + vbCrLf
                If StreamType = 1 Then
                    ' RTP header
                    tbRTPStats.Text += "RTP out of order : " + RTP_OutOfOrder.ToString + vbCrLf
                    tbRTPStats.Text += "RTP bad header   : " + RTP_BadHeader.ToString + vbCrLf
                End If
            End If
            Previous_Input_TSPacketsReceived = Input_TSPacketsReceived

            ' DUMP Stats
            If DumpMode = 0 Then
                tbDumpStats.Text = "No dumping setup..."
            Else
                tbDumpStats.Text = ""
                tbDumpStats.Text += "Files written : " + DumpFileNumber.ToString + vbCrLf
                tbDumpStats.Text += "File size     : " + CDbl(DumpSize / (1024 * 1024)).ToString("0.00") + " MB" + vbCrLf
                If IsNothing(DumpTime) = False Then
                    tbDumpStats.Text += "Time ellapsed : " + DateTime.Now.Subtract(DumpTime).TotalSeconds.ToString("0") + " s" + vbCrLf
                End If
                tbDumpStats.Text += "Actual file : " + vbCrLf + DumpActualFile + vbCrLf
            End If

            ' PID infos
            If cbPause.Checked = False Then
                Dim dt As New DataTable()
                dt.Columns.Add("PID", GetType(String))
                dt.Columns.Add("Received", GetType(String))
                If StreamInput = 1 Then
                    dt.Columns.Add("Frequency", GetType(String))
                End If
                dt.Columns.Add("ContinuityError", GetType(Int32))
                dt.Columns.Add("Type", GetType(String))
                dt.Columns.Add("Description", GetType(String))
                dt.Columns.Add("Optional", GetType(String))
                dt.Columns.Add("Scrambled", GetType(Int32))
                dt.Columns.Add("Unscrambled", GetType(Int32))
                dt.Columns.Add("AdaptationField", GetType(Int32))
                dt.Columns.Add("Payload", GetType(Int32))
                dt.Columns.Add("StreamPriority", GetType(Int32))
                dt.Columns.Add("StreamError", GetType(Int32))

                For Each PID As Int32 In Analyzer.TSPIDS.Keys
                    Dim dr As DataRow = dt.NewRow()
                    dr("PID") = "0x" + PID.ToString("X4") + " - " + PID.ToString()
                    dr("Received") = CDbl(100 * CDbl(DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Received) / CDbl(Analyzer.TSReceived)).ToString("00.0") + "% - " + DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Received.ToString()
                    If StreamInput = 1 Then
                        If IsNothing(PreviousPIDs) = False Then
                            If PreviousPIDs.Contains(PID) = True Then
                                dr("Frequency") = CDbl(DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Received) - CDbl(PreviousPIDs(PID))
                            Else
                                dr("Frequency") = "-"
                            End If
                            PreviousPIDs(PID) = CDbl(DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Received)
                        Else
                            dr("Frequency") = "-"
                        End If
                    End If
                    dr("ContinuityError") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).ContinuityError
                    dr("Type") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Type
                    dr("Description") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Description
                    dr("Optional") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Optional
                    dr("Scrambled") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Scrambled
                    dr("Unscrambled") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).Unscrambled
                    dr("AdaptationField") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).HasAdaptationFields
                    dr("Payload") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).HasPayload
                    dr("StreamPriority") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).StreamPriority
                    dr("StreamError") = DirectCast(Analyzer.TSPIDS(PID), AnalyzeTS.InfoPID).StreamError
                    dt.Rows.Add(dr)
                Next
                dt.DefaultView.Sort = "PID ASC"
                dgvPIDS.DataSource = dt
                If StreamInput = 1 Then
                    Dim i As Int32
                    For i = 0 To dgvPIDS.Rows.Count - 1
                        If CStr(dgvPIDS.Rows(i).Cells("Frequency").Value) = "0" Then
                            dgvPIDS.Rows(i).DefaultCellStyle.BackColor = Color.MistyRose
                        ElseIf CStr(dgvPIDS.Rows(i).Cells("Frequency").Value) = "-" Then
                            dgvPIDS.Rows(i).DefaultCellStyle.BackColor = Color.MintCream
                        End If
                    Next
                End If

                ' Messages
                tbMessagesTS.Text = MessageTS
                tbMessagesPAT.Text = MessagePAT
                tbMessagesPMT.Text = MessagePMT

            End If
        Else
            tbTSStats.Text = "Nothing to display..."
            tbMessagesTS.Text = "Nothing to display..."
            tbRTPStats.Text = "Nothing to display..."
            tbDumpStats.Text = "Nothing to display..."
        End If
    End Sub

    Private Sub btStartStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStartStop.Click
        If btStartStop.Text = "Start" Then

            If rbInputFile.Checked = True Then
                If IO.File.Exists(tbFileName.Text) = False Then
                    MessageBox.Show("File " + tbFileName.Text + " doesn't exist!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                    Return
                End If
                StreamInput = 0
                StreamLocation = tbFileName.Text
            End If

            If rbInputMulticast.Checked = True Then
                Try
                    Dim GroupIP As IPAddress = IPAddress.Parse(tbAddress.Text)
                    Dim PortIP As Int32 = Int32.Parse(tbPort.Text)
                    StreamAddress = New IPEndPoint(GroupIP, PortIP)
                Catch ex As Exception
                    MessageBox.Show("Address " + tbAddress.Text + ":" + tbPort.Text + " isn't correct!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                    Return
                End Try
                StreamInput = 1
            End If

            PreviousPIDs = New Hashtable()

            rbRAW.Enabled = False
            rbRTP.Enabled = False
            rbInputFile.Enabled = False
            rbInputMulticast.Enabled = False
            tbAddress.Enabled = False
            tbPort.Enabled = False
            tbFileName.Enabled = False
            tbDumpFilename.Enabled = False
            cbDumpSetting.Enabled = False
            btBrowseInput.Enabled = False
            btBrowseDump.Enabled = False

            If rbRAW.Checked = True Then StreamType = 0
            If rbRTP.Checked = True Then StreamType = 1

            DumpMode = cbDumpSetting.SelectedIndex
            DumpMaxSize = 0
            DumpMaxTime = 0
            Select Case DumpMode
                Case 2
                    DumpMaxTime = 5 * 60
                Case 3
                    DumpMaxTime = 15 * 60
                Case 4
                    DumpMaxTime = 30 * 60
                Case 5
                    DumpMaxTime = 60 * 60
                Case 6
                    DumpMaxSize = 5 * 1024 * 1024
                Case 7
                    DumpMaxSize = 10 * 1024 * 1024
                Case 8
                    DumpMaxSize = 100 * 1024 * 1024
                Case 9
                    DumpMaxSize = 500 * 1024 * 1024
            End Select
            DumpFile = tbDumpFilename.Text.Trim.ToLower.Replace(".ts", "")

            btStartStop.Text = "Stop"
            btStartStop.ForeColor = Color.DarkRed

            MessageTS = ""
            MessagePAT = ""
            MessagePMT = ""

            If IsNothing(Analyzer) = False Then
                RemoveHandler Analyzer.OnNewTSMessage, AddressOf ShowTSMessage
                RemoveHandler Analyzer.OnNewPATMessage, AddressOf ShowPATMessage
                RemoveHandler Analyzer.OnNewPMTMessage, AddressOf ShowPMTMessage
                Analyzer = Nothing
            End If
            Analyzer = New AnalyzeTS()
            AddHandler Analyzer.OnNewTSMessage, AddressOf ShowTSMessage
            AddHandler Analyzer.OnNewPATMessage, AddressOf ShowPATMessage
            AddHandler Analyzer.OnNewPMTMessage, AddressOf ShowPMTMessage
            bwReader.RunWorkerAsync()

            UpdateStats.Enabled = True

        Else

            If StreamInput = 1 AndAlso IsNothing(ListenUdp) = False Then
                ListenUdp.DropMulticastGroup(StreamAddress.Address)
                ListenUdp.Close()
            End If

            btStartStop.Enabled = False
            bwReader.CancelAsync()

        End If
    End Sub

    Private Sub bwReader_DoWork(ByVal sender As System.Object, ByVal e As System.ComponentModel.DoWorkEventArgs) Handles bwReader.DoWork

        DumpStart()

        ' INPUT = FILE/RAW
        If StreamInput = 0 Then

            Dim fi As New IO.FileInfo(StreamLocation)
            Input_TSEstimatedSize = CLng(fi.Length / 188)
            Input_TSPacketsReceived = 0
            Input_BadLength = 0

            Dim EndOfStream As Boolean = False
            Dim fs As New IO.FileStream(StreamLocation, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read, 188, IO.FileOptions.SequentialScan)
            fs.Position = 0
            Dim buffer(187) As Byte
            Dim VolumeRead As Int32
            Dim i As Int32
            Dim SyncFound As Int32
            While bwReader.CancellationPending = False And EndOfStream = False
                VolumeRead = fs.Read(buffer, 0, 188)
                If VolumeRead <> 188 Then
                    ' End of file
                    EndOfStream = True
                Else
                    ' Test if sync byte isn't present at the beginning => problem
                    If buffer(0) <> &H47 Then
                        Input_BadLength += 1
                        ' Searching the next sync byte
                        SyncFound = 200
                        For i = 1 To 187
                            If buffer(i) = &H47 Then
                                If i < SyncFound Then SyncFound = i
                            End If
                        Next
                        ' If sync found, we resync the stream reading, if not, we wait for the next packet
                        If SyncFound < 200 Then
                            fs.Position = fs.Position - 188 + SyncFound
                        End If
                    Else
                        ' Packet OK
                        Input_TSPacketsReceived += 1
                        Analyzer.NewPacket(buffer)
                        DumpSave(buffer)
                    End If
                End If
            End While
            fs.Close()

        End If

        ' INPUT = MULTICAST
        If StreamInput = 1 Then

            Input_TSPacketsReceived = 0
            Input_PacketsReceived = 0
            Input_BadLength = 0
            RTP_OutOfOrder = 0
            RTP_BadHeader = 0

            Dim Buffer As Byte() = CType(Array.CreateInstance(GetType(Byte), 188), Byte())
            Dim FirstRun As Boolean = True
            Dim RTP_Sequence_Old As Int32 = 0
            Dim RTP_Sequence_New As Int32 = 0
            Dim bteReceiveData As Byte()

            ' Setup the receiving socket with allowing sharing 
            ListenUdp = New UdpClient()
            ListenUdp.Client.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, 1)
            ListenUdp.ExclusiveAddressUse = False
            ListenUdp.Client.Bind(New IPEndPoint(IPAddress.Any, StreamAddress.Port))
            ListenUdp.JoinMulticastGroup(StreamAddress.Address)

            ' Source
            Dim ipeSource As IPEndPoint = Nothing
            Input_Source = "Unknown"

            While bwReader.CancellationPending = False

                ' Réception du paquet
                Try
                    bteReceiveData = ListenUdp.Receive(ipeSource)
                    Input_Source = ipeSource.Address.ToString + ":" + ipeSource.Port.ToString
                    Input_PacketsReceived += 1
                Catch ex As Exception
                    bteReceiveData = Nothing
                End Try

                ' Test longueur paquet RTP
                If IsNothing(bteReceiveData) = False Then

                    ' RAW TS stream
                    If StreamType = 0 Then
                        If (bteReceiveData.Length > 0) AndAlso ((bteReceiveData.Length) Mod 188 = 0) Then
                            ' Splitting the TS packets
                            For Nb = 0 To ((bteReceiveData.Length) / 188) - 1
                                Array.Copy(bteReceiveData, CInt(Nb * 188), Buffer, 0, 188)
                                Analyzer.NewPacket(Buffer)
                                DumpSave(Buffer)
                                Input_TSPacketsReceived += 1
                            Next
                        Else
                            ' Bad length
                            Input_BadLength += 1
                        End If
                    End If

                    ' RTP header
                    If StreamType = 1 Then
                        If (bteReceiveData.Length >= 12) AndAlso ((bteReceiveData.Length - 12) Mod 188 = 0) Then
                            ' Test RTP header
                            If bteReceiveData(0) = 128 And bteReceiveData(1) = 33 Then
                                ' Test out of order
                                RTP_Sequence_New = bteReceiveData(2) * 256 + bteReceiveData(3)
                                If FirstRun = False Then
                                    If Not ((RTP_Sequence_Old + 1 = RTP_Sequence_New) Or (RTP_Sequence_Old = 65535 And RTP_Sequence_New = 0)) Then
                                        ' order problem
                                        RTP_OutOfOrder += 1
                                    End If
                                End If
                                RTP_Sequence_Old = RTP_Sequence_New
                            Else
                                ' bad header
                                RTP_BadHeader += 1
                            End If
                            ' Splitting the TS packets
                            For Nb = 0 To ((bteReceiveData.Length - 12) / 188) - 1
                                Array.Copy(bteReceiveData, CInt(12 + Nb * 188), Buffer, 0, 188)
                                Analyzer.NewPacket(Buffer)
                                DumpSave(Buffer)
                                Input_TSPacketsReceived += 1
                            Next
                        Else
                            ' bad length
                            Input_BadLength += 1
                        End If
                    End If

                End If

            End While

        End If

        DumpStop()

    End Sub

    Private Sub bwReader_RunWorkerCompleted(ByVal sender As System.Object, ByVal e As System.ComponentModel.RunWorkerCompletedEventArgs) Handles bwReader.RunWorkerCompleted

        btStartStop.Enabled = True
        btStartStop.Text = "Start"
        btStartStop.ForeColor = Color.DarkGreen

        rbInputFile.Enabled = True
        rbInputMulticast.Enabled = True
        If rbInputMulticast.Checked = True Then
            rbRAW.Enabled = True
            rbRTP.Enabled = True
        End If

        tbAddress.Enabled = True
        tbPort.Enabled = True
        tbFileName.Enabled = True
        tbDumpFilename.Enabled = True
        cbDumpSetting.Enabled = True
        btBrowseInput.Enabled = True
        btBrowseDump.Enabled = True

        UpdateStats.Enabled = False
        UpdateStatsNow()

    End Sub

    Public Sub ShowTSMessage(ByVal Message As String)
        MessageTS += DateTime.Now.ToString("dd/MM/yyyy HH:mm:ss") + " - " + Message + vbCrLf
    End Sub
    Public Sub ShowPATMessage(ByVal Message As String)
        MessagePAT += DateTime.Now.ToString("dd/MM/yyyy HH:mm:ss") + " - " + Message + vbCrLf
    End Sub
    Public Sub ShowPMTMessage(ByVal Message As String)
        MessagePMT += DateTime.Now.ToString("dd/MM/yyyy HH:mm:ss") + " - " + Message + vbCrLf
    End Sub

    Private Sub btBrowse_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btBrowseInput.Click
        'ofd.FileName = tbFileName.Text
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            tbFileName.Text = ofd.FileName
        End If
    End Sub

    Private Sub Main_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        cbDumpSetting.SelectedIndex = 0
    End Sub

    Private Sub btBrowseDump_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btBrowseDump.Click
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            tbDumpFilename.Text = ofd.FileName
        End If
    End Sub

    ' Force RAW ts for file (no RTP encapsulation)
    Private Sub rbInputFile_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles rbInputFile.CheckedChanged
        rbRAW.Enabled = False
        rbRTP.Enabled = False
        rbRAW.Checked = True
        rbRTP.Checked = False
    End Sub
    Private Sub rbInputMulticast_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles rbInputMulticast.CheckedChanged
        rbRAW.Enabled = True
        rbRTP.Enabled = True
    End Sub

    Private Sub DumpStart()
        If DumpMode > 0 Then
            DumpSize = 0
            DumpTime = DateTime.Now
            DumpActualFile = DumpFile + "-" + DateTime.Now.ToString("yyyy-MM-dd-HH-mm-ss") + ".ts"
            DumpFileNumber = 1
            DumpHandle = New IO.FileStream(DumpActualFile, IO.FileMode.Create)
        End If
    End Sub
    Private Sub DumpSave(ByVal buffer As Byte())
        If DumpMode > 0 Then
            DumpSize += buffer.Length
            If (DumpMaxSize > 0 AndAlso DumpSize > DumpMaxSize) Or (DumpMaxTime > 0 AndAlso DateTime.Now.Subtract(DumpTime).TotalSeconds > DumpMaxTime) Then
                DumpHandle.Flush()
                DumpHandle.Close()
                DumpSize = 0
                DumpTime = DateTime.Now
                DumpActualFile = DumpFile + "-" + DateTime.Now.ToString("yyyy-MM-dd-HH-mm-ss") + ".ts"
                DumpFileNumber += 1
                DumpHandle = New IO.FileStream(DumpActualFile, IO.FileMode.Create)
            End If
            DumpHandle.Write(buffer, 0, buffer.Length)
        End If
    End Sub
    Private Sub DumpStop()
        If DumpMode > 0 Then
            DumpHandle.Flush()
            DumpHandle.Close()
        End If
    End Sub

End Class
